#! /bin/sh

prog=`basename $0`
progdir=`dirname $0`
progdir=`cd $progdir >/dev/null && pwd`
progpid=$$

die() {
    exec 1>&2
    echo "$prog fatal error:  $*"
    kill -15 $progpid
    exit 1
}

init() {
    # NewLine, HorizTab, HorizWhite and AllHorizWhite
    #
    test -z "$srcdir" || cd "$srcdir" || exit 1
    srcdir=`pwd`

    nl='
'   ht='	'
    hw="[ $ht]"
    ahw="${hw}*"

    tmpd=`mktemp -d ${TMPDIR:-/tmp}/err-names-XXXXXX`
    test -d "${tmpd}" || {
        tmpd=${TMPDIR:-/tmp}/err-names-$$
        rm -rf ${tmpd}
        mkdir ${tmpd} || die "cannot make temp directory $tmpd"
    }

    tmpf=${tmpd}/temp.c
    trap "rm -rf ${tmpd}" EXIT
    echo '#include <errno.h>' > ${tmpf}

    test -z "$CC" && {
        CC=`command -v gcc 2>/dev/null`
        test -x "$CC" || CC=`command -v cc`
    }

    errno_list=`
      $CC -E ${tmpf} | \
        sed -n -e 's/\.h[^-_a-zA-Z0-9/].*/.h/' \
            -e 's@.*\(/usr/include.*/errno\)@\1@p' | \
        sort -u`

    grep_expr="^${ahw}#${ahw}define${hw}+E[A-Z_0-9]+${ahw}[0-9]"
    sed_cmd="s/^${ahw}#${ahw}define${ahw}//"

    egrep -h "$grep_expr" $errno_list | \
        sed "$sed_cmd" | \
        awk '{ print $1 " " $2 }' | \
        sort -u -n -k2 > ${tmpd}/sorted.txt
}

compute_limits() {
    max_val=`sed -n '$'"s/^[A-Z_0-9]*${ahw}//p" ${tmpd}/sorted.txt`
    max_width=0
    last_val=-1
    offset=0

    # Do not put this on the "done" line or you may create a subshell.
    #
    exec < ${tmpd}/sorted.txt

    while read nm val
    do
        nmwidth=`echo $nm | wc -c`
        test $nmwidth -gt $max_width && max_width=$nmwidth

        if test $last_val -ne $val
        then
            last_val=`expr $last_val + 1`
            while test $last_val -lt $val
            do
                echo -1 >&6
                last_val=`expr $last_val + 1`
            done
            printf '"%s\\0"\n' ${nm} >&7
            echo $offset >&6
            offset=`expr $offset + ${nmwidth}`
        fi
    done 6> ${tmpd}/err-offsets 7> ${tmpd}/string-frags

    exec < /dev/null
    # the "nmwidth"s include the trailing newline.
    max_width=`expr $max_width - 1`
}

gen_header() {
    nm=`command -v columns` 2>/dev/null
    test -x "$nm" && {
        $nm --end=xxx --help >/dev/null 2>&1 || \
            nm=/use-sed-instead
    }
    test -x "$nm" || {
        columns() {
            if test X$3 = X-S,
            then
                sed -e 's/^/    /' -e 's/$/,/' -e '$s/,/ };/' \
                    ${tmpd}/err-offsets
            else
                sed -e 's/^/    /' -e '$s/$/;/' \
                    ${tmpd}/string-frags
            fi
            echo
        }
    }

    static_pfx='static '
    guard=ERRNO_LIMIT
    max_val=`expr $max_val + 1`
    fmtstr="%5d (%-${max_width}s) == %s"
    fmtlen=`echo "$fmtstr" | wc -c`
    touch ${1}.h || die "source directory cannot be written to"

    cat > ${1}.h <<- _EOF_
	/* -*- buffer-read-only: t -*- vi: set ro:
	 *
	 * DO NOT EDIT THIS FILE ${1}.h
	 *
	 * This file has been derived by scanning system headers
	 * on `date`.
	 *
	`echo "${errno_list}" | sed 's/^/ * /'`
	 *
	`sed -n '/\* *Copyright/,/^ \*\// {
		  p
		  /^ \*\//q
		}' errno.c`

	/**
	 * @file ${1}.h
	 *
	 * @brief
	 * The names, values and counts of error numbers.
	 *
	 * This has been tailored to the `uname -s` version
	 * `uname -r` system for the `uname -m` platform.
	 */
	_EOF_

    sed "s@${1}.h@${1}.c@" ${1}.h > ${1}.c

    cat >> ${1}.h <<- _EOF_
	#ifndef ${guard}

	/**
	 * A count of offset entries for the error name offset table.
	 * In other words, an upper limit on the value of error numbers.
	 * Also used as a header multi-inclusion guard.
	 */
	#define ${guard} ${max_val}

	/**
	 * error number display format.  It has been adjusted so that all
	 * the descriptions line up vertically.
	 * The longest error name is ${max_width} characters.
	 */
	#define ERRNO_DISPLAY_FMT "${fmtstr}\n"

	/**
	 * Get the error name out of the table.
	 * The error number is an index into a table of offsets.
	 * The offset is the caracter index where the name starts.
	 *
	 * @param err  the error number
	 * @returns the name corresponding to the error number, or NULL on error.
	 */
	extern char const * errno_name(int err);
	#endif /* ${guard} */

	/* end of ${1}.h */
	_EOF_

    cat >> ${1}.c <<- _EOF_
	#include <stdio.h> /* define NULL */
	#include "${1}.h"

	char const *
	errno_name(int err)
	{
	    /**
	     * The names of the error numbers known on this system.
	     */
	    static char const errno_names[$offset] =
	`columns -I8 --spread=1 --end=";" -i ${tmpd}/string-frags`

	    /**
	     * table of error name offsets
	     */
	    static int const errno_offsets[${guard}] = {
	`columns -I8 --spread=1 -S, --end=" };" -i ${tmpd}/err-offsets`

	    if (err < 0)
	        err = - err;

	    if (err >= ${guard})
	        return NULL;

	    int idx = errno_offsets[err];
	    if (idx < 0)
	        return NULL;

	    return errno_names + idx;
	}

	/* end of ${1}.c */
	_EOF_
}

gen_docs() {
    test -f errno.c || exit 0
    nm=`command -v doxygen`
    test -x "$nm" || exit 0
    test -d doxygen || mkdir doxygen
    doxygen -g - | \
      sed -e '/^#/d' -e "/^${ahw}\$/d" \
        -e '/^FULL_PATH_NAMES /s/=.*/= NO/' \
        -e '/^OPTIMIZE_OUTPUT_FOR_C /s/=.*/= YES/' \
        -e '/^EXTRACT_/s/=.*/= YES/' \
        -e '/^SOURCE_BROWSER /s/=.*/= YES/' \
        -e '/^PROJECT_NAME /s/=.*/= errno/' \
        -e '/^PROJECT_BRIEF /s/=.*/= show error number descriptions/' \
        -e '/^INPUT /s/=.*/= errno.c err-names.h/' > doxygen/Doxyfile

    nm=`doxygen doxygen/Doxyfile 2>&1` || {
        err=$?
        exec 1>&2
        echo "doxygen exited with $err while producing docs"
        echo "$nm"
        exit 1
    }

    for f in *
    do
        test -d "$f" || continue
        case "$f" in
        doxygen ) : ;;
        * ) test -d doxygen/$f && rm -rf doxygen/$f
            mv -f $f doxygen/. ;;
        esac
    done

    test -d ~/public_html || return 0
    test -d doxygen/html  || return 0
    test -d ~/public_html/errno && rm -rf ~/public_html/errno
    cp -frp doxygen/html ~/public_html/errno
}

init
compute_limits
gen_header err-names
gen_docs
exit 0
# end of mk-err-names.sh
